package ru.dbotthepony.mc.otm.container

import it.unimi.dsi.fastutil.objects.ObjectIterators
import net.minecraft.world.Container
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.core.isNotEmpty

class ContainerIterator(private val container: Container) : IContainerIterator {
	private var index = 0
	private var lastIndex = -1

	override fun hasNext(): Boolean {
		while (index < container.containerSize) {
			if (container[index].isNotEmpty) {
				return true
			}

			index++
		}

		return false
	}

	override fun next(): ItemStack {
		if (!hasNext()) {
			throw NoSuchElementException()
		}

		lastIndex = index
		return container[index++]
	}

	override fun remove() {
		if (lastIndex == -1) {
			throw NoSuchElementException()
		}

		container[lastIndex] = ItemStack.EMPTY
		lastIndex = -1
	}

	override fun setChanged() {
		container.setChanged()
	}
}

fun Container.iterator(): IContainerIterator {
	return if (this is MatteryContainer) {
		iterator()
	} else {
		ContainerIterator(this)
	}
}

class FullContainerIterator(val container: Container, initialPos: Int = 0) : MutableIterator<ItemStack>, ObjectIterators.AbstractIndexBasedListIterator<ItemStack>(0, initialPos), IContainerIterator {
	override fun remove(location: Int) {
		pos++
		container[location] = ItemStack.EMPTY
	}

	override fun get(location: Int): ItemStack {
		return container[location]
	}

	override fun getMaxPos(): Int {
		return container.containerSize
	}

	override fun add(location: Int, k: ItemStack?) {
		throw UnsupportedOperationException()
	}

	override fun set(location: Int, k: ItemStack) {
		container[location] = k
	}

	override fun setChanged() {
		container.setChanged()
	}
}

fun Container.fullIterator() = FullContainerIterator(this)
