package ru.dbotthepony.mc.otm.menu.decorative

import net.minecraft.world.SimpleContainer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.block.entity.decorative.PainterBlockEntity
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.core.ISubscriptable
import ru.dbotthepony.mc.otm.core.addAll
import ru.dbotthepony.mc.otm.core.collect.SupplierMap
import ru.dbotthepony.mc.otm.core.collect.filter
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.core.map
import ru.dbotthepony.mc.otm.core.util.CreativeMenuItemComparator
import ru.dbotthepony.mc.otm.core.util.ResourceLocationValueCodec
import ru.dbotthepony.mc.otm.menu.MatteryMenu
import ru.dbotthepony.mc.otm.menu.MatterySlot
import ru.dbotthepony.mc.otm.menu.input.ItemConfigPlayerInput
import ru.dbotthepony.mc.otm.recipe.PainterRecipe
import ru.dbotthepony.mc.otm.registry.MMenus
import ru.dbotthepony.mc.otm.registry.MRecipes
import java.util.*
import java.util.function.IntConsumer
import kotlin.collections.ArrayList

class PainterMenu(
	containerId: Int, inventory: Inventory, tile: PainterBlockEntity? = null
) : MatteryMenu(MMenus.PAINTER, containerId, inventory, tile) {
	val dyeStored = DyeColor.entries.associateWith { dye ->
		mSynchronizer.ComputedIntField({ tile?.dyeStored(dye) ?: 0 }).also { it.addListener(IntConsumer { rescan() }) }
	}

	val dyeStoredDirect = SupplierMap(dyeStored)
	val itemConfig = ItemConfigPlayerInput(this, tile?.config)

	val inputContainer = MatteryContainer(::rescan, 1)
	val outputContainer = MatteryContainer(1)
	private var lastRecipe: PainterRecipe? = null
	var selectedRecipe by mSynchronizer.Field(null, ResourceLocationValueCodec.nullable).also { it.addListener { rescan() } }

	val selectRecipe = PlayerInput(ResourceLocationValueCodec) {
		selectedRecipe = it
	}

	val inputSlot = object : MatterySlot(inputContainer, 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && inventory.player.level().recipeManager.byType(MRecipes.PAINTER).values.any { it.input.test(itemStack) }
		}
	}

	val outputSlot = object : MatterySlot(outputContainer, 0) {
		override fun tryRemove(p_150642_: Int, p_150643_: Int, p_150644_: Player): Optional<ItemStack> {
			rescan()
			return super.tryRemove(p_150642_, p_150643_, p_150644_)
		}

		override fun onTake(p_150645_: Player, p_150646_: ItemStack) {
			if (p_150646_.isNotEmpty) {
				lastRecipe?.dyes?.let { tile?.takeDyes(it) }
				inputContainer.removeItem(0, 1)
			}

			super.onTake(p_150645_, p_150646_)
		}

		override fun mayPlace(itemStack: ItemStack): Boolean {
			return false
		}
	}

	val dyeSlot = object : MatterySlot(tile?.dyeInput ?: SimpleContainer(1), 0) {
		override fun mayPlace(itemStack: ItemStack): Boolean {
			return super.mayPlace(itemStack) && (DyeColor.getColor(itemStack)?.let { dyeStoredDirect[it]!! + PainterBlockEntity.HUE_PER_ITEM <= PainterBlockEntity.MAX_STORAGE } ?: false)
		}
	}

	init {
		addSlot(outputSlot)
		addSlot(dyeSlot)

		addStorageSlot(dyeSlot)
		addStorageSlot(inputSlot)
		mapQuickMoveToInventory(outputSlot)
		mapQuickMoveToInventory(dyeSlot)

		addInventorySlots()
	}

	override fun removed(p_38940_: Player) {
		super.removed(p_38940_)
		clearContainer(p_38940_, inputContainer)
	}

	val listeners = ISubscriptable.Impl<Unit>()
	val possibleRecipes = ArrayList<PainterRecipe>()

	private fun rescan() {
		possibleRecipes.clear()
		possibleRecipes.addAll(inventory.player.level().recipeManager.byType(MRecipes.PAINTER).values.iterator().filter { it.input.test(inputContainer[0]) })
		possibleRecipes.sortWith(CreativeMenuItemComparator.map { it.output.item })
		listeners.accept(Unit)
		if (tile !is PainterBlockEntity) return

		if (inputContainer.isEmpty || selectedRecipe == null) {
			outputContainer.clearContent()
		} else {
			val recipe = inventory.player.level().recipeManager.byType(MRecipes.PAINTER)[selectedRecipe]

			if (recipe == null || !recipe.canCraft(dyeStoredDirect) || !recipe.matches(inputContainer, inventory.player.level())) {
				outputContainer.clearContent()
			} else {
				outputContainer[0] = recipe.assemble(inputContainer, inventory.player.level().registryAccess())
				lastRecipe = recipe
			}
		}
	}
}
