package ru.dbotthepony.mc.otm.recipe

import com.mojang.serialization.Codec
import com.mojang.serialization.codecs.RecordCodecBuilder
import net.minecraft.core.NonNullList
import net.minecraft.core.RegistryAccess
import net.minecraft.data.recipes.FinishedRecipe
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.Container
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.crafting.Ingredient
import net.minecraft.world.item.crafting.Recipe
import net.minecraft.world.item.crafting.RecipeSerializer
import net.minecraft.world.item.crafting.RecipeType
import net.minecraft.world.level.Level
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.block.entity.decorative.PainterBlockEntity
import ru.dbotthepony.mc.otm.container.get
import ru.dbotthepony.mc.otm.core.isActuallyEmpty
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.tagNotNull
import ru.dbotthepony.mc.otm.data.Codec2RecipeSerializer
import ru.dbotthepony.mc.otm.data.IngredientCodec
import ru.dbotthepony.mc.otm.data.PredicatedCodecList
import ru.dbotthepony.mc.otm.data.minRange
import ru.dbotthepony.mc.otm.registry.MRecipes
import java.util.EnumMap
import java.util.function.Predicate

class PainterRecipe(
	private val id: ResourceLocation,
	val input: Ingredient,
	val output: ItemStack,
	val dyes: Map<DyeColor, Int>
) : Recipe<Container> {
	constructor(
		id: ResourceLocation,
		input: Ingredient,
		output: ItemStack,
		dyes: Set<DyeColor>
	) : this(id, input, output, dyes.associateWith { 1 })

	fun canCraft(storedDyes: Map<DyeColor, Int>): Boolean {
		if (isIncomplete) return false
		if (dyes.isEmpty() || dyes.values.none { it > 0 }) return true
		val copy = EnumMap(storedDyes)

		for ((dye, amount) in dyes) {
			for (i in 0 until amount) {
				PainterBlockEntity.mixer(dye).mix(copy)
				if (copy[dye]!! <= 0) return false
				copy[dye] = copy[dye]!! - 1
			}
		}

		return true
	}

	override fun matches(p_44002_: Container, p_44003_: Level): Boolean {
		if (isIncomplete) return false
		return input.test(p_44002_[0])
	}

	override fun isIncomplete(): Boolean {
		return input.isActuallyEmpty || output.isEmpty
	}

	override fun getIngredients(): NonNullList<Ingredient> {
		return NonNullList.of(Ingredient.EMPTY, input)
	}

	override fun assemble(p_44001_: Container, p_267165_: RegistryAccess): ItemStack {
		return output.copy().also { o ->
			p_44001_[0].tag?.let {
				if (o.tag == null) {
					o.tag = it.copy()
				} else {
					for (k in it.allKeys)
						if (k !in o.tagNotNull)
							o.tagNotNull[k] = it[k]!!.copy()
				}
			}
		}
	}

	override fun canCraftInDimensions(p_43999_: Int, p_44000_: Int): Boolean {
		return true
	}

	override fun isSpecial(): Boolean {
		return true
	}

	override fun getResultItem(p_267052_: RegistryAccess): ItemStack {
		return output
	}

	override fun getId(): ResourceLocation {
		return id
	}

	override fun getSerializer(): RecipeSerializer<*> {
		return SERIALIZER
	}

	override fun getType(): RecipeType<*> {
		return MRecipes.PAINTER
	}

	fun toFinished(): FinishedRecipe {
		return SERIALIZER.toFinished(this)
	}

	companion object {
		val SERIALIZER = Codec2RecipeSerializer<PainterRecipe> { context ->
			RecordCodecBuilder.create {
				it.group(
					context.ingredients.fieldOf("input").forGetter(PainterRecipe::input),
					ItemStack.CODEC.fieldOf("output").forGetter(PainterRecipe::output),
					PredicatedCodecList<Map<DyeColor, Int>>(
						DyeColor.CODEC.xmap({ mapOf(it to 1) }, { it.keys.first() }) to Predicate { it.keys.size == 1 && it.values.first() == 1 },
						Codec.list(DyeColor.CODEC).xmap({ it.associateWith { 1 } }, { ArrayList(it.keys) }) to Predicate { it.values.all { it == 1 } },
						Codec.unboundedMap(DyeColor.CODEC, Codec.INT.minRange(1)) to Predicate { true }
					).fieldOf("dyes").forGetter(PainterRecipe::dyes),
				).apply(it) { a, b, c -> PainterRecipe(context.id, a, b, c) }
			}
		}
	}
}
